
#define LOBS_PRIVATE_USE
#include "lobsdims.h"
#include "lobsc.h"
#if   LOBS_SPACE_DIM == 1
#   include "lobs1d.h"
#elif LOBS_SPACE_DIM == 2
#   include "lobs2d.h"
#elif LOBS_SPACE_DIM == 3
#   include "lobs3d.h"
#else
#error "Bad dimension.  Define LOBS_SPACE_DIM to be equal to 1, 2, or 3"
#endif

/*
#define LOGGING_ALLOC 1 
*/
#include "logmemP.h"

#ifdef BLPOOL_BASED
#  error "Use of blpoolP is pre-configured"
#endif

#undef BLPOOL_BASED
#define BLPOOL_BASED 1
#include "blpoolP.h"


/*
 (L) ibrary for (O) verlapping (B) ox (S) earch
   Copyright (C) 1996 Petr Krysl
*/

typedef struct lobs_box_t  {
  double                 REF_LIST (lox, loy, loz);
  double                 REF_LIST (hix, hiy, hiz);
}               lobs_box_t;

typedef struct lobs_pt_t  {
  double                 REF_LIST (x, y, z);
}               lobs_pt_t;

struct lobs_leaf_t;

typedef struct lobs_obj_ref_t  {
  void              *object;
  lobs_box_t         box;
struct lobs_leaf_t  *leaf;
}               lobs_obj_ref_t;

typedef struct lobs_leaf_t  {
  struct lobs_leaf_t         *parent;
  struct lobs_leaf_t         *descend[2];
  lobs_box_t                  box;
  int                         num_of_objs;
  lobs_obj_ref_t            **objs;
}               lobs_leaf_t;

typedef struct lobs_front_end_t  {
  /* Settable parameter:
     Largest number of objects per tree leaf.  Default is 10.
     */
  int                max_num_of_objects_per_leaf;
  /* Settable parameter:
     Stop searching when you find /find_max_num_objects/.
     Default is find_max_num_objects=INT_MAX
      */
  int                find_max_num_objects;
  /* May be set any time:
     Function to call when an internal error occurs in the LOBS library.
     If not set (i.e., if NULL as set when the instance is initialized,
     the library does an exit ()) .
   */
  lobs_error_notify_func  error_notify;
  /* Internal use only, may be inspected:
     Total number of objects inserted into the tree.
     */
  int                     num_of_inserted_objects;
  /* Internal use only, may be inspected:
     Number of objects found.
     */
  int                     num_of_objects_found;
  /* Internal use only:
     Size of the buffer below.
     */
  int                     size_of_objects_found;
  /* Internal use only, may be inspected:
     For each point found, this array records the object found.
     */
  lobs_obj_ref_t        **objects_found;
  /* Internal use only:
     Counter for the iteration of the objects.
     */
  int                     curr_object;
  /* Internal use only:
     Optimization of storage management.
     */
  void                   *free_array;
  /* Internal use only:
     The range on input
   */
  lobs_box_t              range;
  /* Internal use only:
     The point on input
   */
  lobs_pt_t               pt;
  /* Internal use only:
     The root of the tree.
     */
  struct lobs_leaf_t      *root;
#if BLPOOL_BASED
  blpool_p                 or_pool;
  blpool_p                 or_array_pool;
  blpool_p                 leaf_pool;
#endif
}                lobs_front_end_t;




#define LOBS_ERROR(lobs, explain) \
   {                                                                     \
    if (lobs->error_notify == NULL) {                              \
      fprintf (stderr, "Error in the LOBS library (%s, l.%d) : %s\n",     \
              __FILE__, __LINE__, explain);                             \
      exit (1);                                                          \
    }  else  {                                                            \
      lobs->error_notify (__FILE__, __LINE__, explain);            \
    }                                                                    \
  } 


#define LOBS_CHECK_NEW(lobs, ptr) \
    { if (ptr == NULL) { LOBS_ERROR (lobs, "Failed malloc"); }  } 


/* Ordinary loop */
#define FOR(_Counter_, _Up_) \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++) 


#ifndef min
#   define min(A,B) (( (A) < (B)) ? (A) : (B)) 
#endif
#ifndef max
#   define max(A,B) (( (A) > (B)) ? (A) : (B)) 
#endif
#ifndef min_of_3
#   define min_of_3(A,B,C) \
 (( (A) < (B)) ?              \
 (( (C) < (A)) ? (C) : (A)) :  \
 (( (C) < (B)) ? (C) : (B)) \
           ) 
#   define max_of_3(A,B,C) \
 (( (A) > (B)) ?              \
 (( (C) > (A)) ? (C) : (A)) :  \
 (( (C) > (B)) ? (C) : (B)) \
           ) 
#endif



#define LOBS_NEW(TYPE) ((TYPE *) malloc (sizeof (TYPE)) ) 

#define LOBS_NEW_ARRAY(NELT, TYPE) \
 ((TYPE *) calloc ((NELT) , (size_t) sizeof (TYPE)) ) 

#define LOBS_REALLOC_ARRAY(PTR, NEW_SIZE, TYPE) \
 ((TYPE *) realloc (PTR, (size_t) (NEW_SIZE * sizeof (TYPE)) )) 

#define LOBS_FREE(PTR) free (PTR)

#if BLPOOL_BASED /* blpool-based */

#define NEW_OR_ARRAY(lobs)     blpool_get_mem (lobs->or_array_pool)
#define FREE_OR_ARRAY(lobs, a) 

#define NEW_LEAF(lobs)         blpool_get_mem (lobs->leaf_pool)
#define FREE_LEAF(lobs, leaf)  

#define NEW_OR(lobs)           blpool_get_mem (lobs->or_pool)
#define FREE_OR(lobs, or)      

#else  /* malloc-based */

#define NEW_OR_ARRAY(lobs)     LOBS_NEW_ARRAY (lobs->max_num_of_objects_per_leaf, lobs_obj_ref_t *)
#define FREE_OR_ARRAY(lobs, a) LOBS_FREE (a)

#define NEW_LEAF(lobs)         LOBS_NEW (lobs_leaf_t)
#define FREE_LEAF(lobs, leaf)  LOBS_FREE (leaf)

#define NEW_OR(lobs)           LOBS_NEW (lobs_obj_ref_t)
#define FREE_OR(lobs, or)      LOBS_FREE (or)

#endif

#define LOBS_DUPL_STRING(dest, src) \
   {                                                                      \
    if (( (dest) = LOBS_NEW_ARRAY (strlen (src) +1, char)) == NULL) {     \
      LOBS_CHECK_NEW (NULL);                                              \
    }                                                                     \
    strcpy ((dest) , (src));                                              \
  } 


#define INIT_RANGE(BOX, DIR, V) (BOX) .GLUE (lo,DIR) = (BOX) .GLUE (hi,DIR) = V
#define UPDT_RANGE(BOX, DIR, V) \
   {                                                     \
 (BOX) .GLUE (lo,DIR) = min ((BOX) .GLUE (lo,DIR) , V);   \
 (BOX) .GLUE (hi,DIR) = max ((BOX) .GLUE (hi,DIR) , V);   \
  } 
#define RANGES_OVERLAP(BOX1, BOX2, DIR) \
 (( (BOX1) .GLUE (lo,DIR) <= (BOX2) .GLUE (hi,DIR)) &&    \
 ((BOX1) .GLUE (hi,DIR) >= (BOX2) .GLUE (lo,DIR)) ) 
#define COPY_BOX(DEST, SRC) 	 (DEST) = (SRC) 
#if   LOBS_SPACE_DIM == 1
#  define INIT_BOX(BOX, V0, V1, V2) INIT_RANGE (BOX, x, V0) 
#  define UPDT_BOX(BOX, V0, V1, V2) UPDT_RANGE (BOX, x, V0) 
#  define BOXES_OVERLAP(BOX1, BOX2) RANGES_OVERLAP (BOX1, BOX2, x) 
#define PT_INSIDE_BOX(BOX, X, Y, Z) 		\
 ((X >= (BOX) .lox && X <= (BOX) .hix)) 	
#define BOX_VOLUME(BOX) \
 (( (BOX) .hix - (BOX) .lox)) 
#define INTERSECTION_OF_BOXES(BOX1, BOX2, INTERSECTION) \
    {                                                    \
     INTERSECTION.lox = max (BOX1.lox, BOX2.lox);       \
     INTERSECTION.hix = min (BOX1.hix, BOX2.hix);       \
   } 
#elif LOBS_SPACE_DIM == 2
#  define INIT_BOX(BOX, V0, V1, V2) INIT_RANGE (BOX, x, V0); INIT_RANGE (BOX, y, V1) 
#  define UPDT_BOX(BOX, V0, V1, V2) UPDT_RANGE (BOX, x, V0); UPDT_RANGE (BOX, y, V1) 
#  define BOXES_OVERLAP(BOX1, BOX2) \
 (RANGES_OVERLAP (BOX1, BOX2, x) && \
    RANGES_OVERLAP (BOX1, BOX2, y)) 
#define PT_INSIDE_BOX(BOX, X, Y, Z) 		\
 ((X >= (BOX) .lox && X <= (BOX) .hix) 	\
    && (Y >= (BOX) .loy && Y <= (BOX) .hiy)) 	
#define BOX_VOLUME(BOX) \
 (( (BOX) .hix - (BOX) .lox) *		\
 ((BOX) .hiy - (BOX) .loy)) 
#define INTERSECTION_OF_BOXES(BOX1, BOX2, INTERSECTION) \
    {                                                    \
     INTERSECTION.lox = max (BOX1.lox, BOX2.lox);       \
     INTERSECTION.hix = min (BOX1.hix, BOX2.hix);       \
     INTERSECTION.loy = max (BOX1.loy, BOX2.loy);       \
     INTERSECTION.hiy = min (BOX1.hiy, BOX2.hiy);       \
   } 
#elif LOBS_SPACE_DIM == 3
#  define INIT_BOX(BOX, V0, V1, V2) INIT_RANGE (BOX, x, V0); INIT_RANGE (BOX, y, V1); INIT_RANGE (BOX, z, V2) 
#  define UPDT_BOX(BOX, V0, V1, V2) UPDT_RANGE (BOX, x, V0); UPDT_RANGE (BOX, y, V1); UPDT_RANGE (BOX, z, V2) 
#  define BOXES_OVERLAP(BOX1, BOX2) \
 (RANGES_OVERLAP (BOX1, BOX2, x) && \
    RANGES_OVERLAP (BOX1, BOX2, y) && \
    RANGES_OVERLAP (BOX1, BOX2, z)) 
#define PT_INSIDE_BOX(BOX, X, Y, Z) 		\
 ((X >= (BOX) .lox && X <= (BOX) .hix) 	\
    && (Y >= (BOX) .loy && Y <= (BOX) .hiy) 	\
    && (Z >= (BOX) .loz && Z <= (BOX) .hiz)) 	
#define BOX_VOLUME(BOX) \
 (( (BOX) .hix - (BOX) .lox) *		\
 ((BOX) .hiy - (BOX) .loy) *		\
 ((BOX) .hiz - (BOX) .loz)) 
#define INTERSECTION_OF_BOXES(BOX1, BOX2, INTERSECTION) \
    {                                                    \
     INTERSECTION.lox = max (BOX1.lox, BOX2.lox);       \
     INTERSECTION.hix = min (BOX1.hix, BOX2.hix);       \
     INTERSECTION.loy = max (BOX1.loy, BOX2.loy);       \
     INTERSECTION.hiy = min (BOX1.hiy, BOX2.hiy);       \
     INTERSECTION.loz = max (BOX1.loz, BOX2.loz);       \
     INTERSECTION.hiz = min (BOX1.hiz, BOX2.hiz);       \
   } 
#endif
#define MERGE_BOXES(DESTBOX, BOX1, BOX2) 	                \
   {						                \
    COPY_BOX (DESTBOX, BOX1);		                        \
    UPDT_BOX (DESTBOX, (BOX2) .lox, (BOX2) .loy, (BOX2) .loz); 	\
    UPDT_BOX (DESTBOX, (BOX2) .hix, (BOX2) .hiy, (BOX2) .hiz);	\
  } 
#define BOX_INSIDE_BOX(MASTERBOX, BOX) 	            	         \
 (PT_INSIDE_BOX (MASTERBOX, (BOX) .lox, (BOX) .loy, (BOX) .loz) && \
    PT_INSIDE_BOX (MASTERBOX, (BOX) .hix, (BOX) .hiy, (BOX) .hiz)) 









#define APPEND_OUTPUT(oref)                                     \
   {                                                            \
    if (lobs->num_of_objects_found                              \
        >= lobs->size_of_objects_found) {                       \
      lobs->size_of_objects_found *= 2;                         \
      lobs->objects_found                                       \
        = LOBS_REALLOC_ARRAY (lobs->objects_found,              \
                             lobs->size_of_objects_found,       \
                             lobs_obj_ref_t *);                 \
      LOBS_CHECK_NEW (lobs, lobs->objects_found);               \
                                                                \
    }                                                           \
    lobs->objects_found[lobs->num_of_objects_found] = oref;     \
    lobs->num_of_objects_found++;                               \
  } 


static int
check_leaf (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  int c, n = leaf->num_of_objs;
  
  c = 0;
  while (c < n) {
    if (BOXES_OVERLAP (lobs->range, leaf->objs[c]->box)) {
      APPEND_OUTPUT (leaf->objs[c]);
      if (   lobs->num_of_objects_found
          == lobs->find_max_num_objects) 
        return FALSE; /* stop searching */
    } 
    c++;
  } 
  return TRUE; /* continue searching */
} 




static int
check_leaf_pt (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  int c, n = leaf->num_of_objs;
  
  c = 0;
  while (c < n) {
    if (PT_INSIDE_BOX (leaf->objs[c]->box,
                       lobs->pt.x, lobs->pt.y, lobs->pt.z)) {
      APPEND_OUTPUT (leaf->objs[c]);
      if (   lobs->num_of_objects_found
          == lobs->find_max_num_objects) 
        return FALSE; /* stop searching */
    } 
    c++;
  } 
  return TRUE; /* continue searching */
} 





static lobs_obj_ref_t **
new_objs_array (lobs_front_end lobs) 
{
  if (lobs->free_array) {
    lobs_obj_ref_t **a = lobs->free_array;
    lobs->free_array = NULL;
    return a;
  }  else  {
    lobs_obj_ref_t **a;
    a = NEW_OR_ARRAY (lobs);
    LOBS_CHECK_NEW (lobs, a);
    return a;
  } 
} 

static void
free_objs_array (lobs_front_end lobs, lobs_obj_ref_t **a) 
{
  if (lobs->free_array) {
    FREE_OR_ARRAY (lobs, a);
  }  else  {
    lobs->free_array = a;
  } 
} 


static int
descend_check_leaf (lobs_front_end lobs, lobs_leaf_t *leaf);
static int
descend_check_leaf_pt (lobs_front_end lobs, lobs_leaf_t *leaf);



/*
   This function searches the tree for all the objects whose
   bounding-boxes overlap the box specified on input.  The box
   on input may be a single point, of course. 
   */
int
lobs_search (lobs_front_end lobs,
             REF_LIST (double llx,
                      double lly,
                      double llz) ,
             REF_LIST (double urx,
                      double ury,
                      double urz)) 
{
  INIT_BOX (lobs->range, llx, lly, llz);
  UPDT_BOX (lobs->range, urx, ury, urz);
  
  lobs->num_of_objects_found = 0;

  if (lobs->root) 
    descend_check_leaf (lobs, lobs->root);
  
  return TRUE;
} 




/*
   This function searches the tree for all the objects whose
   bounding-boxes overlap the point specified on input.  The
   function is equivalent to lobs_search () , where the input
   box degenerates to a point.
   */
int
lobs_search_for_boxes_overlapping_pt (lobs_front_end lobs,
                                      REF_LIST (double ptx,
                                               double pty,
                                               double ptz)) 
{
  REF_LIST (lobs->pt.x = ptx, lobs->pt.y = pty, lobs->pt.z = ptz);
  
  lobs->num_of_objects_found = 0;

  if (lobs->root) 
    descend_check_leaf_pt (lobs, lobs->root);
  
  return TRUE;
} 









static lobs_leaf_t *
new_leaf (lobs_front_end lobs, lobs_box_t *box, lobs_leaf_t *parent) 
{
  lobs_leaf_t *l;

  l = NEW_LEAF (lobs);
  LOBS_CHECK_NEW (lobs, l);

  l->parent      = parent;
  l->descend[0]  = NULL;
  l->descend[1]  = NULL;
  COPY_BOX (l->box, *box);
  l->num_of_objs = 0;
  l->objs        = NULL;

  return l;
} 

static int
store (lobs_front_end lobs, lobs_leaf_t *leaf, lobs_obj_ref_t *oref) 
{
  if (leaf->objs == NULL) {
    leaf->objs = new_objs_array (lobs);
    leaf->num_of_objs = 0;
  } 
  if (leaf->num_of_objs >= lobs->max_num_of_objects_per_leaf) {
    return FALSE;
  }  else  {
    leaf->objs[leaf->num_of_objs] = oref;
    leaf->num_of_objs++;
    oref->leaf                    = leaf;
    return TRUE;
  } 
} 

static int
to_which_leaf (lobs_leaf_t *leaf0, lobs_leaf_t *leaf1, lobs_box_t *box) 
{
  double volnewbox0, volnewbox1;
  lobs_box_t newbox;

  MERGE_BOXES (newbox, *box, leaf0->box);
  volnewbox0 = BOX_VOLUME (newbox);
  MERGE_BOXES (newbox, *box, leaf1->box);
  volnewbox1 = BOX_VOLUME (newbox);
  if (volnewbox0  > volnewbox1) {
    return 1;
  }  else  {
    return 0;
  } 
} 

static int
split_leaf (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  lobs_box_t b0, b1; 
  int c, n = leaf->num_of_objs, nh;
  int which;
  lobs_box_t newbox;
  int to[2] =  { 0, 0 } ;

  INIT_BOX (b0, leaf->box.lox, leaf->box.loy, leaf->box.loz);
  INIT_BOX (b1, leaf->box.hix, leaf->box.hiy, leaf->box.hiz);
  leaf->descend[0] = new_leaf (lobs, &b0, leaf);
  leaf->descend[1] = new_leaf (lobs, &b1, leaf);

  nh  = n / 2 + 1;
  c   = 0;
  while (c < n) {
    which = to_which_leaf (leaf->descend[0], leaf->descend[1],
                           &leaf->objs[c]->box);
    if (to[which] > nh) {
      which = (which == 0 ? 1 : 0);
    } 
    to[which]++;
    if (!store (lobs, leaf->descend[which], leaf->objs[c])) 
      LOBS_ERROR (lobs, "failed store");
    MERGE_BOXES (newbox, leaf->descend[which]->box, leaf->objs[c]->box);
    COPY_BOX (leaf->descend[which]->box, newbox);
    c++;
  } 
  
  free_objs_array (lobs, leaf->objs);
  leaf->objs = NULL;

  return TRUE;
} 

#define GROW_SELF(leaf, abox)                   \
   {                                            \
    lobs_box_t _newbox;                         \
    MERGE_BOXES (_newbox, abox, leaf->box);     \
    COPY_BOX (leaf->box, _newbox);              \
  } 


static int
extend (lobs_front_end lobs,
        lobs_leaf_t *parent,
        lobs_leaf_t *leaf,
        lobs_obj_ref_t *oref) 
{
  GROW_SELF (leaf, oref->box);

 retry:
  if (leaf->descend[0] == NULL) { /* terminal leaf */
    if (! store (lobs, leaf, oref)) {
      split_leaf (lobs, leaf);
      goto retry;
    } 
  }  else  {
    int which = to_which_leaf (leaf->descend[0], leaf->descend[1], &oref->box);
    
    extend (lobs, leaf, leaf->descend[which], oref);
  } 
  
  return TRUE;
} 


static lobs_obj_ref_t *
new_obj_ref (lobs_front_end lobs, void *object, lobs_box_t *box) 
{
  lobs_obj_ref_t *oref;
  
  oref = NEW_OR (lobs);
  LOBS_CHECK_NEW (lobs, oref);
  
  oref->object = object;
  oref->box    = *box;
  oref->leaf   = NULL;

  return oref;
} 

/*
   The function adds an object /object/ to the search tree.
   The bounding-box of the object is specified by the coordinates
   of two points, /ll/ and /ur/.  The coordinates of these points
   must bound the object.

   Example:  The sphere of radius 1, with a center at  {-2;3;8} 
   can be described by the bounding-box given by the points
    {-3;4;7}  and  {-1;2;9} .  Note that the lower/upper distinction is
   not necessarily preserved in all coordinate directions for the two
   points.
   */
int
lobs_add_object (lobs_front_end_t *lobs,
                 void *object,
                 REF_LIST (double llx,
                          double lly,
                          double llz) ,
                 REF_LIST (double urx,
                          double ury,
                          double urz)) 
{
  lobs_box_t box;
  lobs_obj_ref_t *oref;

  INIT_BOX (box, llx, lly, llz);
  UPDT_BOX (box, urx, ury, urz);
  
  if (lobs->root == NULL) {
    lobs->root = new_leaf (lobs, &box, NULL);
  } 

  lobs->num_of_inserted_objects++;

  oref = new_obj_ref (lobs, object, &box);
  return extend (lobs, NULL, lobs->root, oref);
} 


/*
  Create a new Lobs (point-in-range tree) .
 */
lobs_front_end
lobs_new_instance (lobs_error_notify_func error_notify,
                   int per_leaf
                   ) 
{
  lobs_front_end_t *lobs;

  lobs = LOBS_NEW (lobs_front_end_t);
  LOBS_CHECK_NEW (lobs, lobs);

  lobs->max_num_of_objects_per_leaf = per_leaf;
  lobs->find_max_num_objects        = INT_MAX;
  lobs->error_notify                = error_notify;
  lobs->num_of_inserted_objects     = 0;
  lobs->num_of_objects_found        = 0;
#if   LOBS_SPACE_DIM == 1
  lobs->size_of_objects_found       =   5;
#elif LOBS_SPACE_DIM == 2
  lobs->size_of_objects_found       =  25;
#elif LOBS_SPACE_DIM == 3
  lobs->size_of_objects_found       = 125;
#endif  
  lobs->objects_found
    = LOBS_NEW_ARRAY (lobs->size_of_objects_found, lobs_obj_ref_t *);
  LOBS_CHECK_NEW (lobs, lobs->objects_found);
  lobs->free_array                  = NULL;
  lobs->root                        = NULL;
#if BLPOOL_BASED
  lobs->or_pool                     = blpool_new (sizeof (lobs_obj_ref_t));
  lobs->or_array_pool               = blpool_new (sizeof (lobs_obj_ref_t *)
                                                  * lobs->max_num_of_objects_per_leaf);
  lobs->leaf_pool                   = blpool_new (sizeof (lobs_leaf_t));
#endif
  
  return lobs;
} 


/*
   Set the maximal number of retrieved objects.
 */
int
lobs_set_find_max_num_objects (lobs_front_end lobs, int find_max_num_objects) 
{
  lobs->find_max_num_objects = find_max_num_objects;
  return TRUE;
} 


static int
print_leaf (lobs_leaf_t *leaf, FILE *fp, int level);
static int
descend_print_leaf (lobs_leaf_t *leaf, FILE *fp, int level);


/*
   Print the structure of the LOBS tree to a stream specified.
 */
int
lobs_print (lobs_front_end_t *lobs, FILE *fp) 
{
  int nt;
  
  fprintf (fp, "Lobs structure:\n");
  nt = descend_print_leaf (lobs->root, fp, 1);
  fprintf (fp, "   total of inserted objects %d\n",
	  lobs->num_of_inserted_objects);
  fprintf (fp, "   total of actually stored objects %d\n", nt);

  return TRUE;
} 


static int
descend_print_leaf (lobs_leaf_t *leaf, FILE *fp, int level) 
{
  int nt;
  
  nt = print_leaf (leaf, fp, level);
  if (leaf->descend[0]) 
    nt += descend_print_leaf (leaf->descend[0], fp, level+1);
  if (leaf->descend[1]) 
    nt += descend_print_leaf (leaf->descend[1], fp, level+1);

  return nt;
} 



static int
print_leaf (lobs_leaf_t *leaf, FILE *fp, int level) 
{
  int nt = 0;
  
  fprintf (fp, "leaf %p\n", leaf);
  fprintf (fp, "   level   = %d\n", level);
#if LOBS_SPACE_DIM == 1
  fprintf (fp, "   box     =  { {%g} ,  {%g} } \n",
	  leaf->box.lox,
          leaf->box.hix);
#elif LOBS_SPACE_DIM == 2
  fprintf (fp, "   box     =  { {%g, %g} ,  {%g, %g} } \n",
	  leaf->box.lox,
	  leaf->box.loy,
	  leaf->box.hix,
	  leaf->box.hiy);
#elif LOBS_SPACE_DIM == 3
  fprintf (fp, "   box     =  { {%g, %g, %g} ,  {%g, %g, %g} } \n",
	  leaf->box.lox,
	  leaf->box.loy,
	  leaf->box.loz,
	  leaf->box.hix,
	  leaf->box.hiy,
	  leaf->box.hiz);
#endif
  if (leaf->objs) {
    int c;
    lobs_obj_ref_t *oref;
    
    nt = leaf->num_of_objs;
    fprintf (fp, "   num_of_objects = %d\n", nt);
    c = 0;
    while (c < nt) {
      oref = leaf->objs[c];
#if LOBS_SPACE_DIM == 1
      fprintf (fp, "   objbox  =  { {%g} ,  {%g} } \n",
              oref->box.lox,
              oref->box.hix);
#elif LOBS_SPACE_DIM == 2
      fprintf (fp, "   objbox  =  { {%g, %g} ,  {%g, %g} } \n",
              oref->box.lox,
              oref->box.loy,
              oref->box.hix,
              oref->box.hiy);
#elif LOBS_SPACE_DIM == 3
      fprintf (fp, "   objbox  =  { {%g, %g, %g} ,  {%g, %g, %g} } \n",
              oref->box.lox,
              oref->box.loy,
              oref->box.loz,
              oref->box.hix,
              oref->box.hiy,
              oref->box.hiz);
#endif
      c++;
    } 
  }  else
    fprintf (fp, "   descend =  {%p, %p} \n", leaf->descend[0], leaf->descend[1]);

  return nt;
} 









void *
lobs_first_overlapping_object (lobs_front_end_t *lobs) 
{
  lobs->curr_object = 0;
  return (lobs->num_of_objects_found > 0
          ? lobs->objects_found[lobs->curr_object]->object
          : NULL);
} 



void *
lobs_next_overlapping_object (lobs_front_end_t *lobs) 
{
  lobs->curr_object++;
  return (lobs->num_of_objects_found > lobs->curr_object
          ? lobs->objects_found[lobs->curr_object]->object
          : NULL);
} 



static int
descend_check_leaf (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  if (leaf->objs) {
    if (BOXES_OVERLAP (lobs->range, leaf->box)) 
      return check_leaf (lobs, leaf);
  }  else  {
    if (leaf->descend[0])
      if (BOXES_OVERLAP (lobs->range, leaf->descend[0]->box)) 
        if (!descend_check_leaf (lobs, leaf->descend[0])) return FALSE;
    if (leaf->descend[1])
      if (BOXES_OVERLAP (lobs->range, leaf->descend[1]->box)) 
        if (!descend_check_leaf (lobs, leaf->descend[1])) return FALSE;
  } 
  return TRUE;
} 




static int
descend_check_leaf_pt (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  if (leaf->objs) {
    if (PT_INSIDE_BOX (leaf->box, lobs->pt.x, lobs->pt.y, lobs->pt.z)) 
      return check_leaf_pt (lobs, leaf);
  }  else  {
    if (PT_INSIDE_BOX (leaf->descend[0]->box, lobs->pt.x, lobs->pt.y, lobs->pt.z)) 
      if (!descend_check_leaf_pt (lobs, leaf->descend[0])) return FALSE;
    if (PT_INSIDE_BOX (leaf->descend[1]->box, lobs->pt.x, lobs->pt.y, lobs->pt.z)) 
      if (!descend_check_leaf_pt (lobs, leaf->descend[1])) return FALSE;
  } 
  return TRUE;
} 



static void
free_leaf (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  if (leaf->objs) {
    int i, n = leaf->num_of_objs;
    for (i = 0; i < n; i++) 
      FREE_OR (lobs, leaf->objs[i]);
    FREE_OR_ARRAY (lobs, leaf->objs); 
  } 
  FREE_LEAF (lobs, leaf);
} 

static void
descend_free_leaf (lobs_front_end lobs, lobs_leaf_t *leaf) 
{
  if (leaf->descend[0]) 
    descend_free_leaf (lobs, leaf->descend[0]);
  if (leaf->descend[1]) 
    descend_free_leaf (lobs, leaf->descend[1]);
  free_leaf (lobs, leaf);
} 


/*
   Free all memory used by the LOBS tree.  The instance of the
   search tree may not be
   used subsequently.
 */
lobs_front_end
lobs_free_instance (lobs_front_end_t *lobs /* front end to destroy */
                    ) 
{
#if BLPOOL_BASED
  blpool_free (lobs->or_pool);
  blpool_free (lobs->or_array_pool);
  blpool_free (lobs->leaf_pool);
#else
  if (lobs->root) 
    descend_free_leaf (lobs, lobs->root);
#endif
  LOBS_FREE (lobs->objects_found);
#if BLPOOL_BASED
#else
  if (lobs->free_array) 
    LOBS_FREE (lobs->free_array);
#endif
  LOBS_FREE (lobs);
  return NULL;
} 

/*
   Recycle instance.
   The stored objects are discarded, and if the implementation supports this,
   the lobs structure returned can still be used by filling in new objects.
   On the other hand, for the malloc-based implementation, this function
   is identical to lobs_free_instance(), and NULL is returned.
 */
lobs_front_end
lobs_recycle_instance (lobs_front_end lobs)
{
#if BLPOOL_BASED
  blpool_recycle (lobs->or_pool);
  blpool_recycle (lobs->or_array_pool);
  blpool_recycle (lobs->leaf_pool);
  lobs->root = NULL;
  return lobs;
#else
  return lobs_free_instance (lobs);
#endif
}

/*
   Get the number of objects found.
   Get the number of objects found to overlap a given box (point).
 */
int
lobs_num_of_objects_found (lobs_front_end lobs)
{
  return lobs->num_of_objects_found;
}



static int
adjust_leaf (lobs_front_end lobs, lobs_leaf_t *leaf, void *object, lobs_box_t *box)
{
  int c, n = leaf->num_of_objs;

  if (leaf->objs) {
    c = 0;
    while (c < n) {
      if (leaf->objs[c]->object == object) {
        leaf->objs[c]->box = *box;
        return TRUE;
      }
      c++;
    }
  }
  return FALSE;
} 

static int
adjust (lobs_front_end lobs,
        lobs_leaf_t *parent,
        lobs_leaf_t *leaf,
        void *object,
        lobs_box_t *box) 
{
  int outcome;
  
  if (leaf->descend[0] == NULL) { /* terminal leaf */
    outcome = adjust_leaf (lobs, leaf, object, box); /* object not found */
    if (outcome) { GROW_SELF (leaf, (*box)); }
    return outcome;
  }  else  {
    int which = to_which_leaf (leaf->descend[0], leaf->descend[1], box);
    outcome = adjust (lobs, leaf, leaf->descend[which], object, box);
    if (outcome) { GROW_SELF (leaf, (*box)); }
    else  outcome = adjust (lobs, leaf, leaf->descend[(which + 1) % 2], object, box);
    return outcome;
  } 
} 


/*
   The function moves an object /object/ stored previously in the search tree.
   The bounding-box of the object is specified by the coordinates
   of two points, /ll/ and /ur/.  The coordinates of these points
   must bound the object.

   Example:  The sphere of radius 1, with a center at  {-2;3;8} 
   can be described by the bounding-box given by the points
    {-3;4;7}  and  {-1;2;9} .  Note that the lower/upper distinction is
   not necessarily preserved in all coordinate directions for the two
   points.
   */
int
lobs_move_object (lobs_front_end lobs,
                  void *object,
                  REF_LIST (double old_llx,
                            double old_lly,
                            double old_llz) ,
                  REF_LIST (double old_urx,
                            double old_ury,
                            double old_urz),
                  REF_LIST (double llx,
                            double lly,
                            double llz) ,
                  REF_LIST (double urx,
                            double ury,
                            double urz)) 
{
  lobs_box_t box;
  int i, n;
  lobs_obj_ref_t *oref;
  lobs_leaf_t *leaf;

  if (lobs->root == NULL) {
    return FALSE; /* root empty; there could not possibly be this object in it */
  } 

  if (!lobs_search (lobs,
                    REF_LIST (old_llx, old_lly, old_llz) ,
                    REF_LIST ( old_urx, old_ury, old_urz)))
    return FALSE;
  
  i = 0; n = lobs->num_of_objects_found; oref = NULL;
  while (i < n) {
    if (lobs->objects_found[i]->object == object) {
      oref = lobs->objects_found[i];
      break;
    }
    i++;
  }

  if (!oref || !oref->leaf)
    return FALSE;
  
  INIT_BOX (box, llx, lly, llz);
  UPDT_BOX (box, urx, ury, urz);

  leaf = oref->leaf;
  oref->box = box; /* update object box */
  while (leaf) {
    GROW_SELF (leaf, box);
    leaf = leaf->parent;
  }
  return TRUE;
} 

