#include <stdio.h>
#include <time.h>

#include "randP.h"
#include "timeP.h"
#include "getoptP.h"
#include "g3d.h"
#define LOGGING_ALLOC 1
#include "logmemP.h"



#ifndef TIMING_MODE
#   define TIMING_MODE   TRUE
#endif
#   define TEST_DIM 3
#   define TEST_BOX 0

#define NOBJS    1000
#define NSRCHS   3
#define NROUNDS  1

#include "lobs3d.h"


typedef struct obj {
  G3D_box_t  box;
}              obj;

static double xscalelo = -1;
static double xscalehi = 1;

static double yscalelo = 0.3;
static double yscalehi = 0.9;

static double zscalelo = 0; 
static double zscalehi = 1; 


static int nobjs       = NOBJS;
static int nrounds     = NROUNDS;
static int pr_scale    = NSRCHS / 10;
static int per_leaf    = 10;
static int inner_loops = 10;
static int move        = FALSE;


static obj *objs;


static double cumulative_lobs_time = 0;

static int
errexit (char *s)
{
  fprintf(stderr, "Error: %s\n", s);
  exit(1);
  return 1;
}


static void 
process_options (int argc, char **argv)
{
  GET_OPT_ARG(argc, argv, "-nobjs", int, nobjs, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-nrounds", int, nrounds, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-per_leaf", int, per_leaf, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-move", int, move, {;},
              "number");
}



static lobs_3d_front_end lobs = NULL;

#define LOBS_NEW_ARRAY(NELT, TYPE) \
                      ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))


/* Ordinary loop */
#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

static void
next_obj (int i, double bsize)
{
  G3D_loc_t loc;
  
  loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
  loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
  loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  G3D_INIT_BOX (objs[i].box, loc);
  G3D_INFLATE_BOX (objs[i].box, bsize);
}

static void
move_obj (int i, double bsize, double msize)
{
  G3D_loc_t loc;
  
  loc.x = (objs[i].box.upper.x + objs[i].box.lower.x) / 2 + msize * rand_0_to_1();
  loc.y = (objs[i].box.upper.y + objs[i].box.lower.y) / 2 + msize * rand_0_to_1();
  loc.z = (objs[i].box.upper.z + objs[i].box.lower.z) / 2 + msize * rand_0_to_1();
  G3D_INIT_BOX (objs[i].box, loc);
  G3D_INFLATE_BOX (objs[i].box, bsize);
}

static double cum_ut;

static void
generate (void)
{
  double bsize = ((xscalehi - xscalelo) +
                  (yscalehi - yscalelo) +
                  (zscalehi - zscalelo)) / nobjs / nobjs / 2;
  double msize = bsize / pow ((double) nobjs, (double) 1/TEST_DIM);
  int i;
  double real_time, usr_time, systm_time, sut;
  G3D_box_t oldbox;

  ckit_get_times (&real_time, &usr_time, &systm_time);
  sut = usr_time;
 
  if (!lobs)
    lobs = lobs_3d_new_instance (NULL, per_leaf);
  
  if (move) {
    FOR(i, nobjs) {
      oldbox = objs[i].box;
      move_obj (i, bsize, msize);
      if (!lobs_3d_move_object (lobs, &objs[i],
                                oldbox.lower.x, oldbox.lower.y, oldbox.lower.z,
                                oldbox.upper.x, oldbox.upper.y, oldbox.lower.z,
                                objs[i].box.lower.x, objs[i].box.lower.y, objs[i].box.lower.z,
                                objs[i].box.upper.x, objs[i].box.upper.y, objs[i].box.upper.z))
        errexit ("failed move");
    }
  } else {
    FOR(i, nobjs) {
      next_obj (i, bsize);
      lobs_3d_add_object (lobs, &objs[i],
                          objs[i].box.lower.x, objs[i].box.lower.y, objs[i].box.lower.z,
                          objs[i].box.upper.x, objs[i].box.upper.y, objs[i].box.upper.z);
    }
  }
  lobs = lobs_3d_recycle_instance (lobs);

  ckit_get_times (&real_time, &usr_time, &systm_time);
  cum_ut += usr_time - sut;
}

static void
round (int r)
{
  r++;
  if (r / 100 * 100 == r) {
    fprintf (stderr, "round %d\n", r);
    system ("ps -al | grep lobsmem");
  }
  generate ();
}

static void
run (void)
{
  int r;

  objs = LOBS_NEW_ARRAY (nobjs, obj);
  if (objs == NULL) errexit("malloc failed");

  lobs = NULL;

  cum_ut = 0;
  for (r = 0; r < nrounds; r++) {
    round (r);
  }
  fprintf (stderr, "%g sec\n", cum_ut);
  free (objs);
  logging_stats (stderr, TRUE);
}


int  
main (int argc, char **argv)
{
  int i;
  
  process_options (argc, argv);
  
  fprintf (stderr, "Timing %dD library allocation/deallocation: %d objects, %d rounds, move %d\n",
           TEST_DIM, nobjs, nrounds, move);
  
  run ();

  fprintf (stderr, "Success\n");

  return 0;
}





