#include <stdio.h>
#include <time.h>

#include "randP.h"
#include "timeP.h"
#include "getoptP.h"
#include "g3d.h"
#define LOGGING_ALLOC 1
#include "logmemP.h"



#   define TIMING_MODE   TRUE
#   define TEST_DIM 3
#   define TEST_BOX 0

#define NOBJS    1000
#define NSRCHS   50
#define NROUNDS  1

#include "lobs3d.h"


typedef struct obj {
  G3D_box_t  box;
}              obj;

static double xscalelo = -1;
static double xscalehi = 1;

static double yscalelo = 0.3;
static double yscalehi = 0.9;

static double zscalelo = 0; 
static double zscalehi = 1; 


static int nobjs       = NOBJS;
static int nsearches   = NSRCHS;
static int nrounds     = NROUNDS;
static int pr_scale    = NSRCHS / 10;
static int per_leaf    = 10;
static int inner_loops = 10;


static obj *objs;


static double cumulative_lobs_time = 0;

static int
errexit (char *s)
{
  fprintf(stderr, "Error: %s\n", s);
  exit(1);
  return 1;
}


static void 
process_options (int argc, char **argv)
{
  GET_OPT_ARG(argc, argv, "-nobjs", int, nobjs, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-nrounds", int, nrounds, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-nsearches", int, nsearches, {;},
              "number");

  GET_OPT_ARG(argc, argv, "-per_leaf", int, per_leaf, {;},
              "number");
}



static lobs_3d_front_end_t  *lobs = NULL;

#define LOBS_NEW_ARRAY(NELT, TYPE) \
                      ((TYPE *) calloc((NELT), (size_t) sizeof(TYPE)))


/* Ordinary loop */
#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

static void
next_obj (int i)
{
  G3D_loc_t loc;
  double bsize = ((xscalehi - xscalelo) +
                  (yscalehi - yscalelo) +
                  (zscalehi - zscalelo)) / nobjs / nobjs / 2;
  
  loc.x = xscalelo + (xscalehi - xscalelo) * rand_0_to_1();
  loc.y = yscalelo + (yscalehi - yscalelo) * rand_0_to_1();
  loc.z = zscalelo + (zscalehi - zscalelo) * rand_0_to_1();
  G3D_INIT_BOX (objs[i].box, loc);
  G3D_INFLATE_BOX (objs[i].box, bsize);
}


static void
generate (void)
{
  int i;

  if (!lobs)
    lobs = lobs_3d_new_instance (NULL, per_leaf);
  
  FOR(i, nobjs) {
    next_obj(i);
    lobs_3d_add_object (lobs, &objs[i],
                        objs[i].box.lower.x,
                        objs[i].box.lower.y,
                        objs[i].box.lower.z,
                        objs[i].box.upper.x,
                        objs[i].box.upper.y,
                        objs[i].box.upper.z);
  }
}

static void
round (int r)
{
  r++;
  if (r / 100 * 100 == r)
    fprintf (stderr, "round %d\n", r);
  lobs = NULL;
  generate ();
  lobs = lobs_3d_recycle_instance (lobs);
}

static void
run (void)
{
  double real_time, usr_time, systm_time, sut;
  int r;

  objs = LOBS_NEW_ARRAY (nobjs, obj);
  if (objs == NULL) errexit("malloc failed");

  ckit_get_times (&real_time, &usr_time, &systm_time);
  sut = usr_time;
  for (r = 0; r < nrounds; r++) {
    round (r);
  }
  ckit_get_times (&real_time, &usr_time, &systm_time);
  fprintf (stderr, "%g sec\n", usr_time - sut);
  free (objs);
  logging_stats (stderr, TRUE);
}


int  
main (int argc, char **argv)
{
  int i;
  
  process_options (argc, argv);
  
  fprintf (stderr, "Timing %dD library: %d objects, %d searches\n",
           TEST_DIM, nobjs, nsearches * inner_loops);
  
  run ();

  fprintf (stderr, "Success\n");

  return 0;
}





